# PyDeco.
#
# Author: Paul Roujansky <paul@roujansky.eu>
# Last Change: June 28, 2019
# URL: https://github.com/paulroujansky/pydeco

"""The main module of the `pydeco` package."""

# Library modules
from .decorator import Decorator, MethodsDecorator
from .utils.misc import PYTHON_VERSION

# Semi-standard module versioning.
__version__ = '0.362'
